from django.shortcuts import get_list_or_404, get_object_or_404, render_to_response
from quran.models import *

def index(request, template_name='quran/index.html'):
    suras = get_list_or_404(Sura)
    return render_to_response(template_name, {'suras': suras})

def sura(request, sura_number, template_name='quran/sura.html'):
    try:
        sura = get_object_or_404(Sura, number=sura_number)
        ayas = sura.ayas.all()
    except Sura.DoesNotExist:
        raise Http404
    return render_to_response(template_name, {'sura': sura, 'ayas': ayas})

def aya(request, sura_number, aya_number, template_name='quran/aya.html'):
    try:
        sura = get_object_or_404(Sura, number=sura_number)
        aya = get_object_or_404(Aya, sura=sura, number=aya_number)
        words = aya.words.all()
    except Sura.DoesNotExist, Aya.DoesNotExist:
        raise Http404
    return render_to_response(template_name, {'sura': sura, 'aya': aya, 'words': words})

def word(request, sura_number, aya_number, word_number, template_name='quran/word.html'):
    aya = get_object_or_404(Aya, sura=sura_number, number=aya_number)
    word = get_object_or_404(Word, aya=aya, number=word_number)
    root = word.root
    return render_to_response(template_name, {'word': word, 'aya': aya, 'root': root})

def lemma(request, lemma_id, template_name='quran/lemma.html'):
    lemma = get_object_or_404(Lemma, pk=lemma_id)
    root = lemma.root
    words = lemma.word_set.all()
    ayas = lemma.ayas.distinct()
    return render_to_response(template_name, {'lemma': lemma, 'root': root, 'words': words, 'ayas': ayas})

def root(request, root_id, template_name='quran/root.html'):
    root = get_object_or_404(Root, pk=root_id)
    lemmas = root.lemmas.all()
    ayas = root.ayas.distinct()
    return render_to_response(template_name, {'root': root, 'lemmas': lemmas, 'ayas': ayas})

def root_index(request, template_name='quran/root_index.html'):
    roots = Root.objects.all().order_by('letters')
    return render_to_response(template_name, {'roots': roots})


import doctest
from quran import buckwalter
from django.test import TestCase

class BuckwalterTest(TestCase):
    def test_buckwalter(self):
        """
        Test the buckwalter.py library.
        """
        doctest.testmod(buckwalter)


_unicode_to_buckwalter = {
u"\u0621": "'", # Hamza
u"\u0622": "|", # Alif + MaddahAbove (not used in Tanzil)
u"\u0623": ">", # Alif + HamzaAbove
u"\u0624": "&", # Waw + HamzaAbove
u"\u0625": "<", # Alif + HamzaBelow
u"\u0626": "}", # Ya + HamzaAbove
u"\u0627": "A", # Alif
u"\u0628": "b", # Ba
u"\u0629": "p", # TaMarbuta
u"\u062A": "t", # Ta
u"\u062B": "v", # Tha
u"\u062C": "j", # Jeem
u"\u062D": "H", # HHa
u"\u062E": "x", # Kha
u"\u062F": "d", # Dal
u"\u0630": "*", # Thal
u"\u0631": "r", # Ra
u"\u0632": "z", # Zain
u"\u0633": "s", # Seen
u"\u0634": "$", # Sheen
u"\u0635": "S", # Sad
u"\u0636": "D", # DDad
u"\u0637": "T", # TTa
u"\u0638": "Z", # DTha
u"\u0639": "E", # Ain
u"\u063A": "g", # Ghain
u"\u0640": "_", # Tatweel
u"\u0641": "f", # Fa
u"\u0642": "q", # Qaf
u"\u0643": "k", # Kaf
u"\u0644": "l", # Lam
u"\u0645": "m", # Meem
u"\u0646": "n", # Noon
u"\u0647": "h", # Ha
u"\u0648": "w", # Waw
u"\u0649": "Y", # AlifMaksura
u"\u064A": "y", # Ya
u"\u064B": "F", # Fathatan
u"\u064C": "N", # Dammatan
u"\u064D": "K", # Kasratan
u"\u064E": "a", # Fatha
u"\u064F": "u", # Damma
u"\u0650": "i", # Kasra
u"\u0651": "~", # Shadda
u"\u0652": "o", # Sukun
u"\u0653": "^", # Maddah (extended)
u"\u0654": "#", # HamzaAbove (extended)
u"\u0670": "`", # AlifKhanjareeya
u"\u0671": "{", # Alif + HamzaWasl
u"\u067E": "P", # Peh (not Quranic)
u"\u0686": "J", # Tcheh (not Quranic)
u"\u06A4": "V", # Veh (not Quranic)
u"\u06AF": "G", # Gaf (not Quranic)
u"\u06DC": ":", # SmallHighSeen (extended)
u"\u06DF": "@", # SmallHighRoundedZero (extended)
u"\u06E0": "\"", # SmallHighUprightRectangularZero (extended)
u"\u06E2": "[", # SmallHighMeemIsolatedForm (extended)
u"\u06E3": ";", # SmallLowSeen (extended)
u"\u06E5": ",", # SmallWaw (extended)
u"\u06E6": ".", # SmallYa (extended)
u"\u06E8": "!", # SmallHighNoon (extended)
u"\u06EA": "-", # EmptyCentreLowStop (extended)
u"\u06EB": "+", # EmptyCentreHighStop (extended)
u"\u06EC": "%", # RoundedHighStopWithFilledCentre (extended)
u"\u06ED": "]", # SmallLowMeem (extended)
u" ": " ",
}

_buckwalter_to_unicode = {}
for (u, bw) in _unicode_to_buckwalter.iteritems():
    _buckwalter_to_unicode[bw] = u;


def buckwalter_to_unicode(str):
    """
    >>> buckwalter_to_unicode('yaHoyaY`')
    u'\u064a\u064e\u062d\u0652\u064a\u064e\u0649\u0670'
    """
    ret = u""
    for c in str:
        ret += _buckwalter_to_unicode[c]
    return ret

def unicode_to_buckwalter(str):
    """
    >>> unicode_to_buckwalter(u'\u064a\u064e\u062d\u0652\u064a\u064e\u0649\u0670')
    'yaHoyaY`'
    """
    ret = ""
    for c in str:
        ret += _unicode_to_buckwalter[c]
    return ret


from django import template
from quran.models import *

register = template.Library()

@register.filter(name='translate')
def translate(value, arg=Translation.get(id=1)):
    """
    Display the translation for the given aya
    """
    return TranslatedAya.get(aya=aya, translation=translation)


from django.db import models

from quran.buckwalter import *


class Sura(models.Model):
    """Sura (chapter) of the Quran"""

    REVELATION_CHOICES = (
        ('Meccan', 'Meccan'),
        ('Medinan', 'Medinan'),
    )

    number = models.IntegerField(primary_key=True, verbose_name='Sura Number')
    name = models.CharField(max_length=50, verbose_name='Sura Name')
    tname = models.CharField(max_length=50, verbose_name='English Transliterated Name')
    ename = models.CharField(max_length=50, verbose_name='English Name')
    order = models.IntegerField(verbose_name='Revelation Order')
    type = models.CharField(max_length=7, choices=REVELATION_CHOICES, verbose_name='')
    rukus = models.IntegerField(verbose_name='Number of Rukus')
    bismillah = models.CharField(max_length=50, blank=True, verbose_name='Bismillah')

    class Meta:
        ordering = ['number']

    def __str__(self):
        return self.tname

    def __unicode__(self):
        return self.name


class Aya(models.Model):
    """Aya (verse) of the Quran"""

    number = models.IntegerField(verbose_name='Aya Number')
    sura = models.ForeignKey(Sura, related_name='ayas', db_index=True)
    text = models.TextField(blank=False)

    class Meta:
        unique_together = (('number', 'sura'))
        ordering = ['sura', 'number']

    def __str__(self):
        return unicode_to_buckwalter(self.text)

    def __unicode__(self):
        return self.text


class QuranTranslation(models.Model):
    """Metadata relating to a translation of the Quran"""
    name = models.CharField(blank=False, max_length=50)
    translator = models.CharField(blank=False, max_length=50)
    source_name = models.CharField(blank=False, max_length=50)
    source_url = models.URLField(blank=False)

    def __unicode__(self):
        return self.name


class TranslatedAya(models.Model):
    """Translation of an aya"""
    sura = models.ForeignKey(Sura, related_name='translations', db_index=True)
    aya = models.ForeignKey(Aya, related_name='translations', db_index=True)
    translation = models.ForeignKey(QuranTranslation, db_index=True)
    text = models.TextField(blank=False)

    class Meta:
        unique_together = (('aya', 'translation'))
        ordering = ['aya']

    def __unicode__(self):
        return self.text


class Root(models.Model):
    """Root word"""

    letters = models.CharField(max_length=10, unique=True, db_index=True) # to my knowledge, there is no root with more than 7 letters
    ayas = models.ManyToManyField(Aya, through='Word')

    def __str__(self):
        return unicode_to_buckwalter(self.letters)

    def __unicode__(self):
        return ' '.join(self.letters)


class Lemma(models.Model):
    """Distinct Arabic word (lemma) in the Quran"""
    token = models.CharField(max_length=50, unique=True, db_index=True)
    root = models.ForeignKey(Root, null=True, related_name='lemmas', db_index=True)
    ayas = models.ManyToManyField(Aya, through='Word')

    class Meta:
        ordering = ['token']

    def __str__(self):
        return unicode_to_buckwalter(self.token)

    def __unicode__(self):
        return self.token

class Word(models.Model):
    """Arabic word in the Quran"""

    sura = models.ForeignKey(Sura, related_name='words', db_index=True)
    aya = models.ForeignKey(Aya, related_name='words', db_index=True)
    number = models.IntegerField()
    token = models.CharField(max_length=50, db_index=True)
    root = models.ForeignKey(Root, null=True, related_name='words', db_index=True)
    lemma = models.ForeignKey(Lemma, db_index=True)

    class Meta:
        unique_together = (('aya', 'number'))
        ordering = ['number']

    def __str__(self):
        return unicode_to_buckwalter(self.token)

    def __unicode__(self):
        return self.token


from django.conf.urls.defaults import *

urlpatterns = patterns('quran.views',
    (r'^$', 'index'),
    (r'^(?P<sura_number>\d+)/$', 'sura'),
    (r'^(?P<sura_number>\d+)/(?P<aya_number>\d+)/$', 'aya'),
    (r'^(?P<sura_number>\d+)/(?P<aya_number>\d+)/(?P<word_number>\d+)/$', 'word'),

    (r'^lemma/(?P<lemma_id>\d+)/$', 'lemma'),
    (r'^root/(?P<root_id>\d+)/$', 'root'),
    (r'^root/$', 'root_index'),
)


import os
from setuptools import setup


# Compile the list of packages available, because distutils doesn't have
# an easy way to do this.
packages, data_files = [], []
root_dir = os.path.dirname(__file__)
if root_dir:
    os.chdir(root_dir)

for dirpath, dirnames, filenames in os.walk('quran'):
    # Ignore dirnames that start with '.'
    for i, dirname in enumerate(dirnames):
        if dirname.startswith('.'): del dirnames[i]
    if '__init__.py' in filenames:
        pkg = dirpath.replace(os.path.sep, '.')
        if os.path.altsep:
            pkg = pkg.replace(os.path.altsep, '.')
        packages.append(pkg)
    elif filenames:
        prefix = dirpath[6:] # Strip "quran/" or "quran\"
        for f in filenames:
            data_files.append(os.path.join(prefix, f))


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "django-quran",
    version = "0.1",
    author = "Idris Mokhtarzada",
    author_email = "idris@umd.edu",
    description = ("Quranic models and helpers for use in Django projects"),
    license = "BSD",
    keywords = "django quran islam arabic",
    url = "http://github.com/idris/django-quran",
    package_dir={'quran': 'quran'},
    packages=packages,
    package_data={'quran': data_files},
    long_description=read('README.md'),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
)


from django.core.management.base import NoArgsCommand
from quran.data import *

class Command(NoArgsCommand):
    help = "Load initial Quran data."

    def handle_noargs(self, **options):
        if Aya.objects.count() > 0:
            print 'The quran database must be empty before running quran_loaddata. Running tests.'
            test_data(verbosity=options['verbosity'])
            return

        print "----- importing quran data (Tanzil) -----"
        import_quran()

        print "----- done importing quran data (Tanzil). starting translations -----"
        import_translations()

        print "----- done importing translations. starting morphology -----"
        import_morphology()

        print "----- done importing morphology. running tests -----"
        test_data(verbosity=options['verbosity'])


import re
import unittest
from os import path
from xml.dom.minidom import parse, parseString
from django.db import transaction

from quran.models import *
from quran.buckwalter import *


def path_to(fn):
    return path.join(path.dirname(__file__), fn)

@transaction.commit_on_success
def import_quran():
    d = parse(path_to('tanzil/quran-data.xml'))
    d2 = parse(path_to('tanzil/quran-uthmani.xml'))
    suras = d.getElementsByTagName('sura')
    for s in suras:
        index = int(s.getAttribute('index'))
        ayas = s.getAttribute('ayas')
        start = int(s.getAttribute('start'))
        name = s.getAttribute('name')
        tname = s.getAttribute('tname')
        ename = s.getAttribute('ename')
        type = s.getAttribute('type')
        order = int(s.getAttribute('order'))
        rukus = int(s.getAttribute('rukus'))
        sura_model = Sura(number=index, name=name, tname=tname, ename=ename, type=type, order=order, rukus=rukus)

        sura = d2.getElementsByTagName('sura')[index - 1]
        assert int(sura.getAttribute('index')) == sura_model.number
        sura_model.save()

        ayas = sura.getElementsByTagName('aya')
        bismillah = ayas[0].getAttribute('bismillah')
        for aya in ayas:
            index = int(aya.getAttribute('index'))
            text = aya.getAttribute('text')
            aya_model = Aya(sura=sura_model, number=index, text=text)
            aya_model.save()
            print "%d:%d" % (sura_model.number, index)


@transaction.commit_on_success
def import_translation_txt(path, translation):
    print "Importing %s translation" % (translation.name)
    f = open(path)
    ayas = Aya.objects.all()
    for aya in ayas:
        line = f.readline()
        if len(line) <= 1:
            raise Exception('Translation file [%s] ended preemtively on aya %d:%d' % (path, aya.sura_id, aya.number))
        line = line.strip()
        t = TranslatedAya(sura=aya.sura, aya=aya, translation=translation, text=line)
        t.save()
        print "[%s] %d:%d" % (translation.name, aya.sura_id, aya.number)


def import_translations():
    translation = QuranTranslation(name='Yusuf Ali', translator='Abdullah Yusuf Ali', source_name='Zekr.org', source_url='http://zekr.org/resources.html')
    translation.save()
    import_translation_txt(path_to('zekr/yusufali.txt'), translation)

    translation = QuranTranslation(name='Shakir', translator='Mohammad Habib Shakir', source_name='Zekr.org', source_url='http://zekr.org/resources.html')
    translation.save()
    import_translation_txt(path_to('zekr/shakir.txt'), translation)

    translation = QuranTranslation(name='Pickthall', translator='Mohammed Marmaduke William Pickthall', source_name='Zekr.org', source_url='http://zekr.org/resources.html')
    translation.save()
    import_translation_txt(path_to('zekr/pickthall.txt'), translation)


def extract_lem(morphology):
    p = re.compile('LEM:(?P<lem>[^ "]+)')
    m = p.search(morphology)
    r = None
    word = None
    if(m):
        r = buckwalter_to_unicode(m.group('lem'))
    return r


def extract_root(morphology):
    p = re.compile('ROOT:(?P<root>[^ "]+)')
    m = p.search(morphology)
    r = None
    root = None
    if m:
        r = buckwalter_to_unicode(m.group('root'))
    if r:
        try:
            root = Root.objects.get(letters=r)
        except Root.DoesNotExist:
            root = Root(letters=r)
            root.save()
    return root


def import_morphology_xml():
    d = parse(path_to('corpus/quranic-corpus-morphology-0.1.xml'))
    suras = d.getElementsByTagName('chapter')
    for s in suras:
        sura_number = int(s.getAttribute('number'))
        sura = Sura.objects.get(number=sura_number)
        ayas = s.getElementsByTagName('verse')
        for a in ayas:
            aya_number = int(a.getAttribute('number'))
            aya = Aya.objects.get(sura=sura, number=aya_number)
            words = a.getElementsByTagName('word')
            for w in words:
                number = int(w.getAttribute('number'))
                token = w.getAttribute('token')
                morphology = w.getAttribute('morphology')

                lemma = None
                dtoken = token
                root = extract_root(morphology)
                lem = extract_lem(morphology)
                if lem: dtoken = lem

                try:
                    lemma = Lemma.objects.get(token=dtoken)
                except Lemma.DoesNotExist:
                    lemma = Lemma(token=dtoken, root=root)
                    lemma.save()

                word = Word(sura=sura, aya=aya, number=number, token=token, root=root, lemma=lemma)
                word.save()

            print "[morphology] %d:%d" % (sura.number, aya.number)


def import_morphology_txt():
    sura = Sura.objects.get(number=2)
    aya = Aya.objects.get(sura=sura, number=2) # any aya except the first.
    f = open(path_to('corpus/quranic-corpus-morphology-0.1.1.txt'))

    line = f.readline()
    while len(line) > 0:
        parts = line.strip().split('|')
        sura_number = 0
        try:
            sura_number = int(parts[0])
        except ValueError:
            line = f.readline()
            continue
        aya_number = int(parts[1])
        word_number = int(parts[2])
        token = parts[3]
        morphology = parts[4]

        if aya_number is not aya.number:
            if sura_number is not sura.number:
                sura = Sura.objects.get(number=sura_number)
            aya = Aya.objects.get(sura=sura, number=aya_number)
            print "[morphology] %d:%d" % (sura.number, aya.number)

        lemma = None
        dtoken = token
        root = extract_root(morphology)
        lem = extract_lem(morphology)
        if lem: dtoken = lem

        try:
            lemma = Lemma.objects.get(token=dtoken)
        except Lemma.DoesNotExist:
            lemma = Lemma(token=dtoken, root=root)
            lemma.save()

        word = Word(sura=sura, aya=aya, number=word_number, token=token, root=root, lemma=lemma)
        word.save()

        line = f.readline()


def import_morphology():
    return import_morphology_txt()


def test_data(verbosity):
    verbosity = int(verbosity)
    print verbosity
    test_suite = unittest.TestLoader().loadTestsFromTestCase(DataIntegrityTestCase)
    unittest.TextTestRunner(verbosity=verbosity).run(test_suite)


class DataIntegrityTestCase(unittest.TestCase):
    def check_word(self, sura_number, aya_number, word_number, expected_word):
        sura = Sura.objects.get(number=sura_number)
        aya = sura.ayas.get(number=aya_number)
        word = aya.words.get(number=word_number)
        self.assertEquals(word.token, buckwalter_to_unicode(expected_word))

    def test_first_ayas(self):
        """
        Test the first ayas of some suras
        """
        self.check_word(1, 1, 3, u'{lr~aHoma`ni')
        self.check_word(2, 1, 1, u'Al^m^')
        self.check_word(114, 1, 1, u'qulo')

    def test_last_ayas(self):
        """
        Test the last ayas of some suras
        """
        self.check_word(1, 7, 2, u'{l~a*iyna')
        self.check_word(2, 286, 49, u'{loka`firiyna')
        self.check_word(114, 6, 3, u'wa{ln~aAsi')

    def test_yusuf_ali(self):
        """
        Test some ayas against Yusuf Ali
        """
        sura_number = 112
        aya_number = 4
        sura = Sura.objects.get(number=sura_number)
        aya = sura.ayas.get(number=aya_number)
        translation = QuranTranslation.objects.get(name='Yusuf Ali')
        t = aya.translations.get(translation=translation)
        self.assertEquals(t.text, 'And there is none like unto Him.')